from threading import Thread
from inotify_simple import INotify, flags

class InotifyWatch(Thread):
    """
    Watches for inotify notifications when the databsae file changes. This
    sends notification to any observers added to this monitor.
    """
    def __init__(self, dbfile):
        super().__init__()
        self.inotify = INotify()
        self.watch = self.inotify.add_watch(dbfile, flags.MODIFY)
        self.running = True
        self.observers = []
        
    def stop(self):
        self.inotify.rm_watch(self.watch)
        self.running = False
    
    def run(self):
        while self.running:
            for event in self.inotify.read():
                for observer in self.observers:
                    observer.notify()
    
    def add_observer(self, observer):
        self.observers.append(observer)



from xdg import XDG_DATA_HOME, XDG_CONFIG_HOME, XDG_CACHE_HOME
import os

class AppDirectories():
    def __init__(self):
        self.data = os.fspath(XDG_DATA_HOME.joinpath('quikey/'))
        self.config = os.fspath(XDG_CONFIG_HOME.joinpath('quikey/'))
        self.cache = os.fspath(XDG_CACHE_HOME.joinpath('quikey/'))
        os.makedirs(self.data, exist_ok=True)
        os.makedirs(self.config, exist_ok=True)    
        os.makedirs(self.cache, exist_ok=True)




from tinydb import TinyDB, Query
from datetime import datetime

class Database():
    def __init__(self, appDirs, dbFile='phrases.json'):
        self.appDirs = appDirs
        self.dbFile = self.appDirs.data + '/' + dbFile
        # Ensure database file exists
        self.db = TinyDB(self.dbFile)

    def get(self, key):
        phrase = Query()
        phraseDict = self.db.get(phrase.key == key)
        return phraseDict.get('value')

    def put(self, key, value, tags=None):
        now = datetime.utcnow().isoformat()
        phrase = {'key': key, 'value': value, 'tags': tags, 'updated': now}
        self.db.insert(phrase)

    def update(self, key, value, tags=None):
        now = datetime.utcnow().isoformat()
        phrase = Query()
        self.db.update({'key': key, 'value': value, 'updated': now}, phrase.key == key)

    def delete(self, key):
        phrase = Query()
        return self.db.remove(phrase.key == key)

    def all(self):
        return self.db.all()



#!/usr/bin/env python

import click
from terminaltables import AsciiTable
import humanize
import signal
import os

from quikey.models import Database
from quikey.directories import AppDirectories
import subprocess

MARKER = '''
# Everything after this line will be ignored.
# Insert your quikey phrase into this file then save and close.
'''

def get_database():
    appDirs = AppDirectories() # XDG folders
    d = Database(appDirs)
    return d

@click.group()
@click.pass_context
def cli(ctx):
    ctx.obj = {'database': get_database()}
    
@cli.command()
@click.option('--name' ,'-n', required=True, help='Name of quikey phrase to add.')
@click.option('--tag', '-t', multiple=True, help='Optional tags for the phrase. You can specify this option multiple times.')
@click.option('--phrase', '-p', help='The full phrase to add. If this option is not specified then your default editor ($EDITOR) will be used.')
@click.pass_context
def add(ctx,name,phrase,tag):
    db = ctx.obj['database']
    contents = None
    if phrase is not None:
        contents = phrase
    else:
        contents = click.edit('\n\n'+MARKER)
        if contents is not None:
            contents = contents.split(MARKER, 1)[0].rstrip('\n')
        else:
            click.echo('quikey phrase with key of %s not added' % name)
            return
    db.put(name, contents, tag)
    click.echo('quikey phrase with key of %s added.' % name)

@cli.command()
@click.option('--name', '-n', required=True, help='Name of quikey phrase to edit.')
@click.pass_context
def edit(ctx,name):
    db = ctx.obj['database']
    phrase = db.get(name)
    if phrase is not None:
        contents = click.edit(phrase + '\n\n' + MARKER)
        if contents is not None:
            contents = contents.split(MARKER, 1)[0].rstrip('\n')
            db.update(name, contents)
            click.echo('quikey phrase with key of %s updated.' % name)
        else:
            click.echo('quikey phrase with key of %s not updated' % name)
    else:
        click.echo('quikey phrase with key of %s does not exist.' % name)
        

@cli.command()
@click.option('--name', '-n', required=True, help='Name of quikey phrase to remove.')
@click.pass_context
def rm(ctx,name):
    db = ctx.obj['database']
    # TODO: Support multiple.
    if db.delete(name):
        click.echo('quikey phrase with key of %s has been deleted.' % name)
    else:
        click.echo('quikey phrase with key of %s does not exist.' % name)
    
@cli.command()
@click.option('--show-all', is_flag=True)
@click.pass_context
def ls(ctx, show_all):
    db = ctx.obj['database']
    table = [['Name', 'Tags', 'Last Modified', 'Phrase']]
    phrases = db.all()
    for phrase in phrases:
        tags = ', '.join([x for x in phrase.get('tags')])
        v = phrase.get('value')
        value = (v[:40] + '...' if len(v) > 40  and not show_all else v)
        table.append([phrase.get('key'), tags, humanize.naturalday(phrase.get('updated')), value])
    output = AsciiTable(table)
    click.echo(output.table)

@cli.command()
@click.pass_context
def status(ctx):
    subprocess.run(['quikey-daemon', 'status'])
    click.echo("Database location: " + ctx.obj['database'].dbFile)

@cli.command()
def start():
    subprocess.run(['quikey-daemon', 'start'])

@cli.command()
def stop():
    subprocess.run(['quikey-daemon', 'stop'])



from pynput.keyboard import Key, Controller, Listener, KeyCode
import re
from collections import deque

class PhraseHandler:
    """
    Watches for notifications about new incoming phrases. A single 
    PhraseHandler is created for each phrase-key in the database. 
    
    PhraseHandler also "types out" the value matching a phrase-key.
    
    The user input is checked with trailing characters with priority. A regular 
    expression is interanlly created like this: .*<key>. For example, if you 
    have a phrase key called "hello", the regular expression would be: .*hello
    
    If a user types "blahblahhello<Enter>", their phrase would be triggered.
    """
    def __init__(self, key, database):
        self.key = key
        escaped = re.escape(key)
        self.db = database
        self.regex = re.compile('^.*' + escaped + '$')
        self.keyboard = Controller()
        
    def notify(self, incomingkey):
        if self.regex.match(incomingkey):
            phrase = self.db.get(self.key)
            self.backspace(len(self.key)+1)
            self.keyboard.type(phrase)
            return True
        return False
    
    def backspace(self, count):
        i = 0
        while i < count:
            self.keyboard.press(Key.backspace)
            self.keyboard.release(Key.backspace)
            i = i + 1

class Notifier:
    """
    Class composed of PhraseHandlers. Each observer called when new key input
    is ready to be checked against phrase database.
    
    Notifier acquires a global lock while iterating observers. This prevents
    infinite loops when a value being "typed out" contains a phrase-key.
    i.e. we don't want the automated keyboard typing to be picked up by this app.
    """
    def __init__(self, lock):
        self.observers = []
        self.lock = lock
    
    def clear(self):
        self.observers = []
    
    def add(self, observer):
        self.observers.append(observer)
        
    def notify(self, key):
        # Acquire lock to that any output from an observer hanlding the
        # phrase-key do not get piped back into this app.
        self.lock.acquire()        
        for observer in self.observers:
            try:
                if observer.notify(key):
                    # Release lock since one of observers handled the phrase-key.
                    self.lock.release()                    
                    return True
            except:
                # TODO: What types of exceptions would be raised here?
                pass
        # No observers hanlded the phrase-key so go head and release lock.
        self.lock.release()
        return False

class AlphaNumHandler:
    """
    Handles alphanumeric input, appending to key input buffer.
    """
    def verify(self, key):
        return type(key) == KeyCode
    
    def onkey(self, key, keybuff):
        if not self.verify(key):
            return False

        if len(keybuff) == keybuff.maxlen:
            # Pop off front of queue if at max length.
            keybuff.popleft()
        # Push new input onto end of queue.
        if key.char is not None:
            keybuff.append(key.char)
        return True
        
class DeleteHandler:
    """
    Handles delete/backspace input, adjusting key input buffer.
    """
    def verify(self, key):
        return key == Key.backspace
    
    def onkey(self, key, keybuff):
        if not self.verify(key):
            return False
        if len(keybuff) > 0:
            # Pop off end of queue as long as there is an item in queue.
            keybuff.pop()
        return True

class SpaceHandler:
    """
    Special handling for space.
    """
    def onkey(self, key, keybuff):
        if key != Key.space:
            return False
        if len(keybuff) == keybuff.maxlen:
            keybuff.popleft()
        keybuff.append(' ')
        return True

class TriggerPhraseHandler:
    """
    Handles "trigger" keys that tell this app when to look for key 
    substitution phrases.
    """
    def __init__(self, trigger_keys):
        self.triggerkeys = []
        for key in trigger_keys:
            key = key.strip()
            if len(key) > 0 and key in Key.__members__:
                self.triggerkeys.append(Key[key])
    
    def verify(self, key):
        return key in self.triggerkeys
    
    def onkey(self, key):
        return self.verify(key)

class InputHandler:
    """
    Handles keyboard input, calling each key-specific handler.
    """
    def __init__(self, lock, notifier, buffer_size, trigger_keys):
        self.keybuff = deque(maxlen=buffer_size)
        self.lock = lock
        self.notifier = notifier
        self.subhandlers = []
        self.triggerhandler = TriggerPhraseHandler(trigger_keys)

    def dequetostr(self):
        ret = u''
        for x in self.keybuff:
            ret = ret + x
        return ret        

    def __call__(self, key):
        if self.lock.locked():
            # Skip. Something else has acquired lock.
            return 
        # Check if trigger key pressed
        if self.triggerhandler.onkey(key):
            if self.notifier.notify(self.dequetostr()):
                # A phrase was found and typed out. Clear queue and return.
                self.keybuff.clear()
                return
        for handler in self.subhandlers:
            result = handler.onkey(key, self.keybuff)
            if result:
                return
            
    def add_handler(self, handler):
        self.subhandlers.append(handler)
    
def backspace(count):
    keyboard = Controller()

    i = 0
    while i < count:
        keyboard.press(Key.backspace)
        keyboard.release(Key.backspace)
        i = i + 1




#!/usr/bin/env python
from pynput.keyboard import Key, Listener, KeyCode
from threading import Lock
import daemon
import click
import os
import signal
import sys

from quikey.models import Database
from quikey.directories import AppDirectories
from quikey.filewatch import InotifyWatch
from quikey.input import PhraseHandler, Notifier, InputHandler, AlphaNumHandler, DeleteHandler, SpaceHandler

typelock = Lock()

class DatabaseChangeHandler:
    def __init__(self, notifier, database):
        self.notifier = notifier
        self.db = database
        self.init_phrase_handlers()

    def init_phrase_handlers(self):
        self.notifier.clear()
        phrases = self.db.all()
        for phrase in phrases:
            self.notifier.add(PhraseHandler(phrase.get('key'), self.db))
    
    def notify(self, event=None):
        self.init_phrase_handlers()
        
class ShutdownHook:
    def __init__(self, listener, watch):
        self.watch = watch
        self.listener = listener
        
    def __call__(self, signal, frame):
        appDirs = AppDirectories() # XDG folders
        self.watch.stop()
        self.listener.stop()
        delete_pid(appDirs)

def write_pid(appDirs):
    pidfile = appDirs.cache + '/quikey.pid'
    with open(pidfile, 'w') as f:
        f.write(str(os.getpid()))

def read_pid(appDirs):
    pidfile = appDirs.cache + '/quikey.pid'
    try:
        with open(pidfile, 'r') as f:
            return f.read()
    except FileNotFoundError:
        return None

def delete_pid(appDirs):
    pidfile = appDirs.cache + '/quikey.pid'
    os.remove(pidfile)

def main(foreground, buffer_size, trigger_keys):
    # Initialize all components and hook them up.
    appDirs = AppDirectories() # XDG folders
    notifier = Notifier(typelock) # Create the notifier that calls to each phrase handler
    database = Database(appDirs) # Read database in
    dbchange = DatabaseChangeHandler(notifier, database)
    watch = InotifyWatch(database.dbFile) 
    watch.add_observer(dbchange)  # Watch for changes in database outside this process
    i = InputHandler(typelock, notifier, buffer_size, trigger_keys) # Accepts keyboard inputs
    watch.start()
    i.add_handler(DeleteHandler()) # Special behavior for deletes
    i.add_handler(AlphaNumHandler()) # Standard behavior for everything else
    i.add_handler(SpaceHandler())
    write_pid(appDirs) # Store the current pid
    listener = Listener(on_press=i)
    with Listener(on_press=i) as listener: # Continue listening until SIGTERM
        signal.signal(signal.SIGTERM, ShutdownHook(listener, watch))
        signal.signal(signal.SIGINT, ShutdownHook(listener, watch))
        listener.join()

@click.group()
@click.pass_context
def cli(obj):
    pass

@cli.command()
@click.option('--foreground' ,'-f', is_flag=True, required=False, default=False, help='Run the quikey daemon process in foreground.')
@click.option('--buffer-size', '-b', required=False, default=32, help='Size of buffer that stores keystrokes.')
@click.option('--trigger-keys', '-t', multiple=True, required=False, default=['enter', 'space'], help='Trigger keys that indicate the end of a key phrase. The key name should match one from https://pythonhosted.org/pynput/_modules/pynput/keyboard/_base.html#Key')
def start(foreground, buffer_size, trigger_keys):
    appDirs = AppDirectories() # XDG folders
    pid = read_pid(appDirs)
    daemon_log = appDirs.data+'/qkdaemon.log'

    if pid:
        print('Quikey daemon is already running (pid: %s).' % pid)
        exit
    if foreground:
        main(foreground, buffer_size, trigger_keys)
    else:
        daemon_log_f = open(daemon_log, 'w+')
        with daemon.DaemonContext(stdout=daemon_log_f,stderr=daemon_log_f):
            main(foreground, buffer_size, trigger_keys)

@cli.command()
def stop():
    # Send SIGTERM signal
    appDirs = AppDirectories() # XDG folders
    pid = read_pid(appDirs)
    if pid is None:
        print("No Quikey daemon currently running.")
        return
    try:
        os.kill(int(pid), signal.SIGTERM)
    except ProcessLookupError:
        print("No Quikey daemon currently running (tried killing non-existent pid %s)." % pid)
        delete_pid(appDirs)

@cli.command()
def status():
    # Check process existence
    appDirs = AppDirectories() # XDG folders
    pid = read_pid(appDirs)
    if pid is None:
        print ("No Quikey daemon is currently running")
        sys.exit(1)
    try:
        os.kill(int(pid), 0)
    except OSError:
        print ("No Quikey daemon is currently running")
        sys.exit(1)
    else:
        print ("Quikey daemon running (PID: " + pid +")")
        return


#!/usr/bin/env python3
from setuptools import setup, find_packages

VERSION='0.1.0'

setup(
    name='quikey',
    version=VERSION,
    packages=find_packages(),
    author='bostrt',
    entry_points = {
        'console_scripts': [
            'qk=quikey.quikey:cli',
            'quikey-daemon=quikey.qkdaemon:cli'
            ]
    },
    description='A keyboard macro tool.',
    license='',
    url='https://github.com/bostrt/quikey',
    install_requires=[
        'click', 
        'colored',
        'python-daemon',
        'inotify-simple',
        'tinydb',
        'pynput',
        'terminaltables',
        'xdg',
        'humanize'
        ]
)
